/*
 * bcas_cardio.c - PC/SC card I/F for BCAS descrambling
 * Copyright (C) 2009 0p1pp1
 * 
 * Authors:
 *   0p1pp1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include "bcas_cardio.h"

#ifndef CONFIG_PCSC
bcas_card_t *bcas_card_init(void) { return NULL; }
void bcas_card_stop(bcas_card_t * card) {}
void  bcas_ecm_cb(struct _bcas_card *bcas, void *kinfo,
		uint8_t *rbuf, size_t rlen) {}
#else

#include <winscard.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>

#include "mpbswap.h"
#include "mp_msg.h"

#define CARD_SW1_OK 0x90
#define CARD_SW2_OK 0x00

#define CARD_TYPE_GENERAL 0x01

#define SYS_MNG_ID_B14 0x0301
#define SYS_MNG_ID_B15 0x0201

#define CA_SYSTEM_ID_ARIB 0x0005
#define CA_SYSTEM_ID_ARIB_B 0x000A

#define BCAS_MULTI2_ROUND 4

#define CARD_RETCODE_OK 0x2100
#define CARD_RETCODE_GD_TIER 0x0800
#define CARD_RETCODE_GD_PREPPV 0x0400
#define CARD_RETCODE_GD_POSTPPV 0x0200
#define CARD_RETCODE_PV_PREPPV 0x4480
#define CARD_RETCODE_PV_POSTPPV 0x4280

#define CARD_DIR_PWRON_CTL (1 << 0)
#define CARD_DIR_RINGING_DATE (1 << 1)
#define CARD_DIR_DEL_PASSWD (1 << 2)
#define CARD_DIR_RING (1 << 3)
#define CARD_DIR_RETRY_OVER (1 << 5)
#define CARD_DIR_HANGUP_LINE (1 << 6)
#define CARD_DIR_INIT_PARAM (1 << 7)
#define CARD_DIR_CARD_ID (1 << 8)
#define CARD_DIR_CHANGE_CARD (1 << 9)

void *bcas_card_io(void *data);

struct CardInfo
{
  SCARDCONTEXT cxt;
  SCARDHANDLE hcard;
};

/* example ATR of BCAS. note: byte[2],[4],[8-9],[11-12] may vary */
static const BYTE bcas_atr[] = {
  0x3B, 0xF0, 0x12, 0x00, 0xFF, 0x91, 0x81,
  0xB1, 0x7C, 0x45, 0x1F, 0x03, 0x99
};

static const DWORD BCAS_ATR_LEN = sizeof (bcas_atr);

static char *
get_card (SCARDCONTEXT cxt, SCARDHANDLE * h, char * iccname)
{
  LONG ret;
  DWORD len, protocol;
  char *p, *readers;
  DWORD nlen, alen;
  DWORD state;
  BYTE atr[MAX_ATR_SIZE];

  ret = SCardListReaders (cxt, NULL, NULL, &len);
  if (ret != SCARD_S_SUCCESS) {
    mp_msg (MSGT_DEMUX, MSGL_V, "Failed to get the length of the card list: %s\n",
        pcsc_stringify_error (ret));
    return NULL;
  }

  readers = (char *) malloc (len);
  if (!readers) {
    mp_msg (MSGT_DEMUX, MSGL_INFO, "Failed to alloc memory for the card list.\n");
    return NULL;
  }

  ret = SCardListReaders (cxt, NULL, readers, &len);
  if (ret != SCARD_S_SUCCESS) {
    mp_msg (MSGT_DEMUX, MSGL_V, "Failed to get the card list: %s\n",
        pcsc_stringify_error (ret));
    free (readers);
    return NULL;
  }

  for (p = readers; *p; p += strlen (p) + 1) {
    if (iccname &&
        (strlen (p) != strlen (iccname) || strncmp (p, iccname, strlen (p))))
      continue;
    ret = SCardConnect (cxt, p, SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1,
        h, &protocol);
    if (ret != SCARD_S_SUCCESS) {
      mp_msg (MSGT_DEMUX, MSGL_V, "Failed to connect the card [%s]/%s.\n",
          p, pcsc_stringify_error (ret));
      continue;
    }
    alen = sizeof (atr);
    ret = SCardStatus (*h, NULL, &nlen, &state, &protocol, atr, &alen);
    if (ret == SCARD_S_SUCCESS && (state & SCARD_PRESENT) &&
        alen == BCAS_ATR_LEN &&
        atr[0] == bcas_atr[0] && atr[1] == bcas_atr[1]) {
      if (!iccname)
        iccname = strdup (p);
      free (readers);
      return iccname;
    }
  }
  free (readers);
  return NULL;
}

static int
bcas_card_setup (bcas_card_t * card)
{
  struct CardInfo *info = card->io_arg;
  LONG ret;
  DWORD protocol, rlen;
  SCARD_IO_REQUEST rpci;
  uint8_t rspbuf[BCAS_MAX_MSG_LEN];

  char *name;
  uint retry;
  uint16_t cas_id;
  int i;

  static const uint8_t init_cmd[] = { 0x90, 0x30, 0x00, 0x00, 0x00 };
  static const uint8_t cardinfo_cmd[] = { 0x90, 0x32, 0x00, 0x00, 0x00 };

  // card->status = CARD_S_NG;

  /* init card and set some ID infos */
  ret = SCardEstablishContext (SCARD_SCOPE_SYSTEM, NULL, NULL, &info->cxt);
  if (ret != SCARD_S_SUCCESS) {
    mp_msg (MSGT_DEMUX, MSGL_INFO, "Can't connect to pcscd. (%s).\n",
        pcsc_stringify_error (ret));
    return -1;
  }
//  SCardSetTimeout (info->cxt, 3);

  name = get_card (info->cxt, &info->hcard, NULL);
  if (!name) {
    mp_msg (MSGT_DEMUX, MSGL_INFO, "Failed to find an available BCAS card.\n");
    goto bailout;
  }
  /* discard the old name string if any */
  if (card->iccname)
    free (card->iccname);
  card->iccname = name;

  retry = 0;
  while (retry < 3) {
    memcpy (&rpci, SCARD_PCI_T1, sizeof (rpci));
    rlen = sizeof (rspbuf);
    ret = SCardTransmit (info->hcard, SCARD_PCI_T1, init_cmd, sizeof (init_cmd),
        &rpci, rspbuf, &rlen);
    if (ret == SCARD_W_RESET_CARD) {
      ret = SCardReconnect (info->hcard, SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1,
          SCARD_RESET_CARD, &protocol);
    }
    if (ret == SCARD_S_SUCCESS)
      break;
    retry++;
    sleep (1);
  }
  if (retry == 3 || rlen < 59) {
    mp_msg (MSGT_DEMUX, MSGL_V, "Failed to initialize the card.\n");
    goto bailout;
  }
  cas_id = rspbuf[6] << 8 | rspbuf[7];
  if (rspbuf[rlen - 2] != CARD_SW1_OK || rspbuf[rlen - 1] != CARD_SW2_OK ||
      rspbuf[1] < 55 || (rspbuf[4] << 8 | rspbuf[5]) != CARD_RETCODE_OK ||
      (cas_id != CA_SYSTEM_ID_ARIB && cas_id != CA_SYSTEM_ID_ARIB_B) ||
      rspbuf[14] != CARD_TYPE_GENERAL) {
    mp_msg (MSGT_DEMUX, MSGL_V, "Got bad response for the card init command.\n");
    goto bailout;
  }
  card->cas_id = cas_id;

  /* omit checking the card directions(rspbuf[2-3]) and retcode(rspbuf[4-5]) */
  /* omit checking if sys_management_id[](rspbuf[57..]) includes the valid ID */

  /* copy the initial data */
  for (i = 0; i < 8; i++) {
    card->param.k_sys[i] = be2me_32 (*(u32 *) (rspbuf + 16 + i * 4));
  }
  card->param.cbc_init.whole = be2me_64 (*((u64 *) & rspbuf[48]));

  /* redundant. overwritten very soon by the next command */
  memcpy (&card->id[0][2], &rspbuf[8], 6);
  card->num_id = 1;

  /* get the card info */
  retry = 0;
  while (retry < 3) {
    memcpy (&rpci, SCARD_PCI_T1, sizeof (rpci));
    rlen = sizeof (rspbuf);
    ret = SCardTransmit (info->hcard, SCARD_PCI_T1,
        cardinfo_cmd, sizeof (cardinfo_cmd), &rpci, rspbuf, &rlen);
    if (ret == SCARD_W_RESET_CARD) {
      ret = SCardReconnect (info->hcard, SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1,
          SCARD_RESET_CARD, &protocol);
    }
    if (ret == SCARD_S_SUCCESS)
      break;
    retry++;
    sleep (1);
  }
  if (retry == 3 || rlen < 19) {
    mp_msg (MSGT_DEMUX, MSGL_V, "Failed to get info from the the card.\n");
    goto bailout;
  }

  if (rspbuf[rlen - 2] != CARD_SW1_OK || rspbuf[rlen - 1] != CARD_SW2_OK ||
      rspbuf[1] < 15 || (rspbuf[4] << 8 | rspbuf[5]) != CARD_RETCODE_OK ||
      rspbuf[6] < 1 || rspbuf[6] > 8) {
    mp_msg (MSGT_DEMUX, MSGL_V, "Got bad response for the card info command.\n");
    goto bailout;
  }

  /* copy the card IDs */
  card->num_id = rspbuf[6];
  memcpy (card->id, &rspbuf[7], 10 * rspbuf[6]);
  card->status = CARD_S_OK;
  mp_msg (MSGT_DEMUX, MSGL_INFO, "BCAS card init succeeded.\n");
  return 0;

bailout:
  if (info->hcard)
    SCardDisconnect (info->hcard, SCARD_LEAVE_CARD);
  SCardReleaseContext (info->cxt);
  free (card->io_arg);
  card->io_arg = NULL;
  mp_msg (MSGT_DEMUX, MSGL_INFO, "BCAS card not available.\n");
  return -1;
}

void *
bcas_card_io (void *data)
{
  bcas_card_t *card = (bcas_card_t *) data;
  struct CardInfo *info = card->io_arg;
  LONG ret;
  SCARDCONTEXT cxt = info->cxt;
  SCARDHANDLE hcard = info->hcard;
  DWORD protocol, rlen;
  SCARD_IO_REQUEST rpci;

  uint retry = 0;
  uint8_t cmdbuf[BCAS_MAX_MSG_LEN], rspbuf[BCAS_MAX_MSG_LEN];
  volatile int clen;
  void (*cmd_cb)(struct _bcas_card *card, void *data, uint8_t *rbuf, size_t rlen);
  void *cmd_arg;

  mp_msg (MSGT_DEMUX, MSGL_V, "card I/O thread started.\n");

#ifdef BCAS_THREADED_IO
  while (1)
#else
  do
#endif
  {
    pthread_mutex_lock (&card->lock);
    while ((clen = card->cmd_len) == 0) {
      mp_msg (MSGT_DEMUX, MSGL_DBG2, "card I/O thread sleeping.\n");
      pthread_cond_wait (&card->cond, &card->lock);
    }

    mp_msg (MSGT_DEMUX, MSGL_V, "accepted a card I/O request."
            " cmd_len:%d, cmd_cb:%p\n", clen, card->cmd_cb);
    /* copy the args */
    if (clen > 0 && clen <= BCAS_MAX_MSG_LEN)
      memcpy (cmdbuf, card->cmdbuf, clen);
    cmd_cb = card->cmd_cb;
    cmd_arg = card->cmd_arg;
    card->cmd_len = 0;
    pthread_mutex_unlock (&card->lock);

    if (clen < 0)
      break;

    if (clen > BCAS_MAX_MSG_LEN) {
      mp_msg (MSGT_DEMUX, MSGL_INFO, "Too long CARD command (%dB).\n", clen); 
      goto failed;
    }

    memcpy (&rpci, SCARD_PCI_T1, sizeof (rpci));
    rlen = sizeof (rspbuf);
    mp_msg (MSGT_DEMUX, MSGL_DBG2, "pt0 ok.\n");
    ret = SCardTransmit (hcard, SCARD_PCI_T1, cmdbuf, clen, &rpci, rspbuf, &rlen);
    mp_msg (MSGT_DEMUX, MSGL_DBG2, "pt1 ok.\n");
    if (ret == SCARD_W_RESET_CARD) {
      mp_msg (MSGT_DEMUX, MSGL_V, "resetting the card.\n"); 
      ret = SCardReconnect (hcard, SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1,
          SCARD_RESET_CARD, &protocol);
    }

    if (ret != SCARD_S_SUCCESS) {
      mp_msg (MSGT_DEMUX, MSGL_V, "Card command failed.\n"); 
      if (++retry >= 3) {
        mp_msg (MSGT_DEMUX, MSGL_INFO, "Failed to retry %d times.\n", retry);
        goto bailout;
      }
      SCardDisconnect (hcard, SCARD_RESET_CARD);
      sleep (1);
      ret = SCardConnect (cxt, card->iccname, SCARD_SHARE_SHARED,
          SCARD_PROTOCOL_T1, &hcard, &protocol);
      if (ret != SCARD_S_SUCCESS) {
        mp_msg (MSGT_DEMUX, MSGL_INFO, "Failed to re-open the card [%s].\n",
            card->iccname);
        goto bailout;
      } else {
        goto failed;
      }
    }
    retry = 0;

    /* check if failed and update the status */
    if (rlen < 6 ||
        rspbuf[rlen - 2] != CARD_SW1_OK || rspbuf[rlen - 1] != CARD_SW2_OK) {
      mp_msg (MSGT_DEMUX, MSGL_V,
        "Got the bad response from the card. len:%ld, SW1/2:[0x%02X%02X].\n",
        rlen, rspbuf[rlen - 2], rspbuf[rlen - 1]);
      goto bailout;
    }

    /* currently, we ignore the directions from IC cards. just print it. */
    if (rspbuf[2] || rspbuf[3]) {
      mp_msg (MSGT_DEMUX, MSGL_WARN, "Directions from the card.[0x%04x].\n",
        (uint16_t) (rspbuf[2] << 8 | rspbuf[3]));
    }
    /* call the callback func. */
    if (cmd_cb)
      cmd_cb (card, cmd_arg, rspbuf, rlen);

    continue;

failed:
    pthread_mutex_lock (&card->lock);
    if (cmd_arg) {
      struct demulti2_key_info *kinfo = cmd_arg;
      kinfo->status = CMD_S_FAILED;
    }
    pthread_mutex_unlock (&card->lock);
    continue;

bailout:
    pthread_mutex_lock (&card->lock);
    if (cmd_arg) {
      struct demulti2_key_info *kinfo = cmd_arg;
      kinfo->status = CMD_S_FAILED;
    }
    card->status = CARD_S_NG;
    pthread_mutex_unlock (&card->lock);
    break;
  }
#ifndef BCAS_THREADED_IO
  while (0);
#endif

  mp_msg (MSGT_DEMUX, MSGL_DBG2, "Card I/O thread exiting...\n");
  return NULL;
}


#define MAX_ERROR_RETRY 3
static int io_error_count = 0;

/* exported funcs */

void
bcas_ecm_cb (struct _bcas_card *card, void *data, uint8_t *rbuf, size_t rlen)
{
  struct demulti2_key_info *kinfo = data;
  uint16_t retcode;
  uint64_t k_odd, k_even;

  mp_msg (MSGT_DEMUX, MSGL_V, "Got an ECM response.\n");

#define ECM_RSP_UNIT_SIZE 21
  if (rlen < ECM_RSP_UNIT_SIZE + 4 || rbuf[1] != ECM_RSP_UNIT_SIZE) {
    mp_msg (MSGT_DEMUX, MSGL_INFO,
        "Bad ECM response, len:%ld, unit-len:%d.\n", rlen, rbuf[1]);
    goto failed;
  }
  retcode = rbuf[4] << 8 | rbuf[5];
  if (retcode != CARD_RETCODE_GD_TIER &&
      retcode != CARD_RETCODE_GD_PREPPV &&
      retcode != CARD_RETCODE_GD_POSTPPV &&
      retcode != CARD_RETCODE_PV_PREPPV && retcode != CARD_RETCODE_PV_POSTPPV) {
    mp_msg (MSGT_DEMUX, MSGL_WARN,
        "Bad ret code[0x%04x] in a ECM response.\n", retcode);
    goto failed;
  }
  io_error_count = 0;

  k_odd = be2me_64 (*(uint64_t *) & rbuf[6]);
  k_even = be2me_64 (*(uint64_t *) & rbuf[14]);

  pthread_mutex_lock (&card->lock);
  if (k_odd != kinfo->k_scr[0].whole) {
    mp_msg (MSGT_DEMUX, MSGL_V, "ECM k_odd updated.\n");
    kinfo->k_scr[0].whole = k_odd;
    core_schedule (kinfo->wrk[0], card->param.k_sys, &kinfo->k_scr[0].sub);
  }
  if (k_even != kinfo->k_scr[1].whole) {
    mp_msg (MSGT_DEMUX, MSGL_V, "ECM k_even updated.\n");
    kinfo->k_scr[1].whole = k_even;
    core_schedule (kinfo->wrk[1], card->param.k_sys, &kinfo->k_scr[1].sub);
  }
  kinfo->status = CMD_S_OK;
  pthread_mutex_unlock (&card->lock);

  return;

failed:
  pthread_mutex_lock (&card->lock);
  kinfo->status = CMD_S_FAILED;
  if (++io_error_count > MAX_ERROR_RETRY) {
    mp_msg (MSGT_DEMUX, MSGL_WARN, "Card I/O failed too many times.\n");
    card->status = CARD_S_NG;
  }
  pthread_mutex_unlock (&card->lock);
  return;
}


void
bcas_card_stop (bcas_card_t * card)
{
  struct timespec wt1 = {0,300000000};
  if (card == NULL)
    return;

  pthread_mutex_lock(&card->lock);
  card->cmd_len = -1; // tells the thread to quit
  pthread_cond_broadcast(&card->cond);
  pthread_mutex_unlock(&card->lock);
  pthread_cancel(card->thread);
//  pthread_join(card->thread, NULL);
  pthread_detach(card->thread);
  nanosleep(&wt1, NULL);

  mp_msg (MSGT_DEMUX, MSGL_V, "card I/O thread stopped.\n");

  if (card->io_arg) {
    struct CardInfo *info = (struct CardInfo *) card->io_arg;

    if (info->hcard)
      SCardDisconnect (info->hcard, SCARD_LEAVE_CARD);
    SCardReleaseContext (info->cxt);
    free (card->io_arg);
    card->io_arg = NULL;
  }
  card->cmd_len = 0;
  if (card->iccname)
    free (card->iccname);
  free(card);
}

bcas_card_t *
bcas_card_init (void)
{
  bcas_card_t *card = (bcas_card_t *) calloc(1, sizeof(bcas_card_t));

  if (card == NULL)
    return NULL;

  // card->status = CARD_S_NG;  /* card is calloc'ed. */
  card->param.round = BCAS_MULTI2_ROUND;

  card->io_arg = calloc(1, sizeof(struct CardInfo));
  if (card->io_arg != NULL) {
    pthread_mutex_init (&card->lock, NULL);
    pthread_cond_init (&card->cond, NULL);
    if (bcas_card_setup (card)) {
      free (card->io_arg);
      card->io_arg = NULL;
    }
  }
  if (card->io_arg == NULL) {
    free (card);
    return NULL;
  }

  if (pthread_create(&card->thread, NULL, bcas_card_io, card)) {
    mp_msg (MSGT_DEMUX, MSGL_WARN, "Could't create card I/O thread.\n");
    free (card->io_arg);
    free (card);
    return NULL;
  }
  return card;
}

#endif // CONFIG_PCSC
